import argparse
import time
import numpy as np
import loompy
import networkx as nx
from INSISTC_utils import anchor_based_random_walks_forest_fire, create_dicts, read_anchor
import os
import random
from sklearn.decomposition import LatentDirichletAllocation

"""
running example 
python INSIST_main.py --network example_network_mini_Human.txt --loom example.loom --anchor example_anchor_Human.txt 
python INSIST_main.py --network example_network_mini_Human.txt --loom example.loom --anchor example_anchor_Human.txt --running_mode 1
python INSIST_main.py --loom example.loom --running_mode 2

memory requirement: 32GB or above(may consumer more memory depends on network size)

"""

if __name__ == '__main__':
    my_parser = argparse.ArgumentParser(description="INSIST")
    my_parser.add_argument("--running_mode", action="store", type=int, required=False, default=0,
                           help='INSIST running mode (0(default):walk based and gene based; 1:walk based only; 2:gene based only)')
    my_parser.add_argument("--network", action="store", type=str, required=False, default='',
                           help='Path for network(e.g. gene regulatory netowrk), each row in file contains a definition of edges, tab separated.')
    my_parser.add_argument("--anchor", action="store", type=str, required=False, default='',
                           help='Path for pre-defined anchor node, require a text file where each line cotains an anchor')
    my_parser.add_argument("--loom", action="store", type=str, required=True,
                           help='Path for loom file contains Single-Cell Expression Matrix. The row are "Gene" and the column are "CellID"')
    my_parser.add_argument("--out", action="store", type=str, required=False, default='.',
                           help='Path for output file')
    my_parser.add_argument("--n_lda", action="store", type=int, required=False, default=20,
                           help='The number of topics LDA model')
    my_parser.add_argument("--walk_len", action="store", type=int, required=False, default=5,
                           help='The max length of each random walk')
    my_parser.add_argument("--n_walks", action="store", type=int, required=False, default=5,
                           help='The max number of walks per anchor')
    my_parser.add_argument("--random_seed", action="store", type=int, required=False, default=99,
                           help='The random seed')
    my_parser.add_argument("--max_cores", action="store", type=int, required=False, default=-1,
                           help='max core should be ultilized, -1 == use all')
    my_parser.add_argument("--job_title", action="store", type=str, required=False, default=str(int(time.time())),
                           help='job name')
    my_parser.add_argument("--expression_word_count_scale", action="store", type=int, required=False, default=100,
                           help='parameters for gene expression word count scale(default : 100(expression in UMI count))')

    args = my_parser.parse_args()

    random.seed(args.random_seed)

    start_time = time.time()
    print('> INSISTC v1.')
    if args.running_mode in {0, 1}:
        if not args.network:
            raise Exception(
                "Please specify network configuration edge file (--network DIR_TO_FILE) for walk based and gene based running mode")
        if not args.anchor:
            raise Exception(
                "Please specify network configuration anchor file(--anchor DIR_TO_FILE) for walk based and gene based running mode")

    if args.running_mode == 2 and (args.network or args.anchor):
        print('*** Warning ***: --running_mode and --anchor has no effect under running mode 2(gene only)')

    print('> input loom file dir:', args.loom)
    print('> output file dir:', args.out)

    print('> The number of LDA topics:', args.n_lda)

    if args.running_mode == 0:
        print('> running using walk based word only.')

        print('> The max length of each random walk:', args.walk_len)
        print('> The max number of walks per anchor:', args.n_walks)

        print('> reading loom')
        with loompy.connect(args.loom) as ds:
            cell_names = ds.ca.CellID
            exp_cell_genes = ds.ra.Gene  # words
            exp_mtx = ds.view[:, :]
            cell_names = list(cell_names)
            exp_cell_genes = list(exp_cell_genes)
            occur_matrix = []
            for each in exp_mtx:
                occur_matrix.append(np.asarray(list(each)))
            occur_matrix = np.asarray(occur_matrix).T
            print('# debug, cells', len(cell_names))
            print('# debug, genes', len(exp_cell_genes))
            print('# debug, mtx', occur_matrix.shape)
            loom_genes = ds.ra['Gene']

            dict_exp_arr = {}
            for c, arr in zip(cell_names, occur_matrix):
                dict_exp_arr[c] = arr

        print('\n--- Reading network information(depends on network size, this may takes a while) ---')
        dict_qualify_tf_targets = create_dicts(args.network, loom_genes)

        print('\n--- Mapping nodes ---')

        all_in_one = []

        for TF in dict_qualify_tf_targets.keys():
            all_in_one.append(TF)
            all_in_one = all_in_one + dict_qualify_tf_targets[TF]

        all_in_one = np.unique(all_in_one)
        n_nodes = len(all_in_one)
        print('> num of node', n_nodes)
        dict_gene_to_num = {}
        for i, each in enumerate(all_in_one):
            dict_gene_to_num[each] = i
        dict_num_to_gene = {v: k for k, v in dict_gene_to_num.items()}

        print('\n--- Encoding Graph ---')

        g = nx.Graph()
        for i in range(n_nodes):
            g.add_node(i)

        for TF in dict_qualify_tf_targets.keys():
            tf_n = dict_gene_to_num[TF]
            targets = dict_qualify_tf_targets[TF]
            for target in targets:
                tg_n = dict_gene_to_num[target]
                g.add_edge(tf_n, tg_n)

        print('> number of nodes:', g.number_of_nodes())
        print('> number of edges:', g.number_of_edges())

        print('\n--- Anchor Based Random Walk ---')
        anchor_names = read_anchor(args.anchor)
        anchor_nodes = []

        for each in anchor_names:
            if each in dict_gene_to_num:
                anchor_nodes.append(dict_gene_to_num[each])
        print('> num of anchor nodes', len(anchor_nodes))

        random_walks = anchor_based_random_walks_forest_fire(g, anchor_nodes, num_walks=args.n_walks,
                                                             walk_length=args.walk_len)
        with open(args.out + '/Walks' + args.job_title + '.gmt', 'w') as f_out:
            for i, walk in enumerate(random_walks):
                tmp_arr = ['walk_' + str(i), 'na']

                for each in walk:
                    if dict_num_to_gene[each] in loom_genes:
                        tmp_arr.append(dict_num_to_gene[each])
                if len(tmp_arr) > 3:
                    print('\t'.join(tmp_arr), file=f_out)

        print('\n--- Calculate AUCell ---')
        os.system('pyscenic aucell ' +
                  args.loom + ' ' +
                  args.out + '/Walks' + args.job_title + '.gmt' +
                  ' -o ' + args.out + '/AUCell_mtx_' + args.job_title + '.auc.csv')

        occur_matrix = []
        cell_names = []

        words_name = []
        with open(args.out + '/AUCell_mtx_' + args.job_title + '.auc.csv', 'r') as f_in:
            for i, line in enumerate(f_in):
                if i == 0:

                    words_name += line.strip('\n').split(',')[1:] + list(loom_genes)
                    continue
                else:
                    content = line.strip('\n').split(',')
                    cell_name = content[0]
                    cell_names.append(cell_name)
                    aucell_vals = content[1:]
                    aucell_vals = [float(i) for i in aucell_vals]
                    tmp = np.zeros(len(aucell_vals) + len(exp_cell_genes))
                    skip_idx = len(aucell_vals)
                    for idx, val in enumerate(aucell_vals):
                        tmp[idx] = int(val * args.expression_word_count_scale)
                    this_cell_exp_arr = dict_exp_arr[cell_name]

                    # add gene as word
                    tmp[skip_idx:] = this_cell_exp_arr

                    occur_matrix.append(tmp)
        occur_matrix = np.asarray(occur_matrix)
        # print('# debug, occur matrix shape', occur_matrix.shape)
        print('\n--- LDA ---')
        lda = LatentDirichletAllocation(n_components=args.n_lda,
                                        random_state=args.random_seed, learning_method='batch',
                                        n_jobs=args.max_cores)  # online method suppose to be faster

        lda.fit(occur_matrix)
        topic_arr = lda.transform(occur_matrix)
        # pickle.dump(lda, open(args.out + 'lda_model.sav', 'wb'))

        with open(args.out + '/LDA_topic_distribution_matrix_' + args.job_title + '.txt', 'w') as f_out:
            out_header = ['cellID'] + ['topic#' + str(i) for i in range(args.n_lda)]
            print('\t'.join(out_header), file=f_out)
            for c_id, each_vec in zip(cell_names, topic_arr):
                print('\t'.join([c_id] + [str(i) for i in each_vec]), file=f_out)
        print('> Save topics distribution matrix as :',
              args.out + '/LDA_topic_distribution_matrix_' + args.job_title + '.txt')
        print('> time used(s):', time.time() - start_time)

    elif args.running_mode == 1:
        print('> running using walk based word only.')

        with loompy.connect(args.loom) as ds:
            loom_genes = ds.ra['Gene']
            loom_genes = set(loom_genes)
            cell_names = ds.ca.CellID
            exp_cell_genes = ds.ra.Gene  # words
            exp_mtx = ds.view[:, :]
            cell_names = list(cell_names)
            exp_cell_genes = list(exp_cell_genes)
            occur_matrix = []
            for each in exp_mtx:
                occur_matrix.append(np.asarray(list(each)))
            occur_matrix = np.asarray(occur_matrix).T
            # print('# debug, cells', len(cell_names))
            # print('# debug, genes', len(exp_cell_genes))
            # print('# debug, mtx', occur_matrix.shape)

            dict_exp_arr = {}
            for c, arr in zip(cell_names, occur_matrix):
                dict_exp_arr[c] = arr
        print('\n--- Reading network information(depends on network size, this may takes a while) ---')
        dict_qualify_tf_targets = create_dicts(args.network, loom_genes)

        print('\n--- Mapping nodes ---')

        all_in_one = []

        for TF in dict_qualify_tf_targets.keys():
            all_in_one.append(TF)
            all_in_one = all_in_one + dict_qualify_tf_targets[TF]

        all_in_one = np.unique(all_in_one)
        n_nodes = len(all_in_one)
        print('> num of node', n_nodes)
        dict_gene_to_num = {}
        for i, each in enumerate(all_in_one):
            dict_gene_to_num[each] = i
        dict_num_to_gene = {v: k for k, v in dict_gene_to_num.items()}

        print('\n--- Encoding Graph ---')

        g = nx.Graph()
        for i in range(n_nodes):
            g.add_node(i)

        for TF in dict_qualify_tf_targets.keys():
            tf_n = dict_gene_to_num[TF]
            targets = dict_qualify_tf_targets[TF]
            for target in targets:
                tg_n = dict_gene_to_num[target]
                g.add_edge(tf_n, tg_n)

        print('> number of nodes:', g.number_of_nodes())
        print('> number of edges:', g.number_of_edges())

        print('\n--- Anchor Based Random Walk ---')
        anchor_names = read_anchor(args.anchor)
        anchor_nodes = []

        for each in anchor_names:
            if each in dict_gene_to_num:
                anchor_nodes.append(dict_gene_to_num[each])
        print('> num of anchor nodes', len(anchor_nodes))

        random_walks = anchor_based_random_walks_forest_fire(g, anchor_nodes, num_walks=args.n_walks,
                                                             walk_length=args.walk_len)
        with open(args.out + '/Walks' + args.job_title + '.gmt', 'w') as f_out:
            for i, walk in enumerate(random_walks):
                tmp_arr = ['walk_' + str(i), 'na']

                for each in walk:
                    if dict_num_to_gene[each] in loom_genes:
                        tmp_arr.append(dict_num_to_gene[each])
                if len(tmp_arr) > 3:
                    print('\t'.join(tmp_arr), file=f_out)

        print('\n--- Calculate AUCell ---')
        os.system('pyscenic aucell ' +
                  args.loom + ' ' +
                  args.out + '/Walks' + args.job_title + '.gmt' +
                  ' -o ' + args.out + '/AUCell_mtx_' + args.job_title + '.auc.csv')

        print('\n--- Load AUCell result, creating words ---')
        occur_matrix = []
        cell_names = []
        words_name = []
        with open(args.out + '/AUCell_mtx_' + args.job_title + '.auc.csv', 'r') as f_in:
            for i, line in enumerate(f_in):
                if i == 0:
                    words_name += line.strip('\n').split(',')[1:]
                    continue
                else:
                    content = line.strip('\n').split(',')
                    cell_name = content[0]
                    cell_names.append(cell_name)
                    aucell_vals = content[1:]
                    aucell_vals = [float(i) for i in aucell_vals]
                    tmp = np.zeros(len(aucell_vals))  # version 5!!
                    skip_idx = len(aucell_vals)
                    for idx, val in enumerate(aucell_vals):
                        tmp[idx] = int(val * args.expression_word_count_scale)
                    this_cell_exp_arr = dict_exp_arr[cell_name]

                    # add gene as word
                    # tmp[skip_idx:] = this_cell_exp_arr

                    occur_matrix.append(tmp)
        occur_matrix = np.asarray(occur_matrix)
        print('# debug, occur matrix shape', occur_matrix.shape)

        print('\n--- LDA ---')

        lda = LatentDirichletAllocation(n_components=args.n_lda,
                                        random_state=0, learning_method='batch',
                                        n_jobs=-1)  # online method suppose to be faster

        lda.fit(occur_matrix)

        topic_arr = lda.transform(occur_matrix)
        topic_word_matrix = lda.components_
        with open(args.out + '/LDA_topic_distribution_matrix_' + args.job_title + '.txt', 'w') as f_out:
            out_header = ['cellID'] + ['topic#' + str(i) for i in range(args.n_lda)]
            print('\t'.join(out_header), file=f_out)
            for c_id, each_vec in zip(cell_names, topic_arr):
                print('\t'.join([c_id] + [str(i) for i in each_vec]), file=f_out)
        print('> Save topics distribution matrix as :',
              args.out + '/LDA_topic_distribution_matrix_' + args.job_title + '.txt')
        print('> time used(s):', time.time() - start_time)

    elif args.running_mode == 2:
        print('> running using gene based word only.')

        print('> reading loom')
        with loompy.connect(args.loom) as ds:
            cell_names = ds.ca.CellID
            exp_cell_genes = ds.ra.Gene  # words
            exp_mtx = ds.view[:, :]
            cell_names = list(cell_names)
            exp_cell_genes = list(exp_cell_genes)
            occur_matrix = []
            for each in exp_mtx:
                occur_matrix.append(np.asarray(list(each)))
            occur_matrix = np.asmatrix(occur_matrix).T
            loom_genes = ds.ra['Gene']
            words_name = list(loom_genes)

        occur_matrix = np.asarray(occur_matrix)
        # print('# debug, occur matrix shape', occur_matrix.shape)

        print('\n--- LDA ---')

        lda = LatentDirichletAllocation(n_components=args.n_lda,
                                        random_state=0, learning_method='batch',
                                        n_jobs=-1)  # online method suppose to be faster

        lda.fit(occur_matrix)
        # pickle.dump(lda, open(save_dir + 'lda_model' + PARAM_F_NAME_2 + '.sav', 'wb'))

        # print('# debug', lda.transform(occur_matrix[-2:]))
        topic_arr = lda.transform(occur_matrix)
        topic_word_matrix = lda.components_
        with open(args.out + '/LDA_topic_distribution_matrix_' + args.job_title + '.txt', 'w') as f_out:
            out_header = ['cellID'] + ['topic#' + str(i) for i in range(args.n_lda)]
            print('\t'.join(out_header), file=f_out)
            for c_id, each_vec in zip(cell_names, topic_arr):
                print('\t'.join([c_id] + [str(i) for i in each_vec]), file=f_out)
        print('> Save topics distribution matrix as :',
              args.out + '/LDA_topic_distribution_matrix_' + args.job_title + '.txt')
        print('> time used(s):', time.time() - start_time)

    else:
        raise Exception(
            "Wrong running mode assigned. please assign correct running mode. \nINSIST running mode: "
            "\n\t0(default):walk based and gene based; \n\t1:walk based only; \n\t2:gene based only")
